@@ -1,6 +1,6 @@
 /*
     PowerDNS Versatile Database Driven Nameserver
-    Copyright (C) 2003 - 2006  PowerDNS.COM BV
+    Copyright (C) 2003 - 2008  PowerDNS.COM BV
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License version 2 as published 
@@ -16,6 +16,7 @@
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
+#include <boost/algorithm/string.hpp>
 #include "utility.hh"
 #include "syncres.hh"
 #include <iostream>
@@ -33,6 +34,7 @@
 #include "lwres.hh"
 #include "recursor_cache.hh"
 #include "dnsparser.hh"
+#include "dns_random.hh"
 
 extern MemRecursorCache RC;
 
@@ -93,7 +95,9 @@ int SyncRes::beginResolve(const string &qname, const QType &qtype, uint16_t qcla
     return 0;
   }
   
-  if(qclass!=1)
+  if(qclass==0xff)
+    qclass=1;
+  else if(qclass!=1)
     return -1;
   
   set<GetBestNSAnswer> beenthere;
@@ -127,7 +131,9 @@ bool SyncRes::doOOBResolve(const string &qname, const QType &qtype, vector<DNSRe
   
   ret.clear();
   AuthDomain::records_t::const_iterator ziter;
+  bool somedata=false;
   for(ziter=range.first; ziter!=range.second; ++ziter) {
+    somedata=true;
     if(qtype.getCode()==QType::ANY || ziter->qtype==qtype || ziter->qtype.getCode()==QType::CNAME)  // let rest of nameserver do the legwork on this one
       ret.push_back(*ziter);
   }
@@ -136,6 +142,19 @@ bool SyncRes::doOOBResolve(const string &qname, const QType &qtype, vector<DNSRe
     res=0;
     return true;
   }
+  if(somedata) {
+    LOG<<prefix<<qname<<": found record in '"<<authdomain<<"', but nothing of the right type, sending SOA"<<endl;
+    ziter=iter->second.d_records.find(make_tuple(authdomain, QType(QType::SOA)));
+    if(ziter!=iter->second.d_records.end()) {
+      DNSResourceRecord rr=*ziter;
+      rr.d_place=DNSResourceRecord::AUTHORITY;
+      ret.push_back(rr);
+    }
+    else
+      LOG<<prefix<<qname<<": can't find SOA record '"<<authdomain<<"' in our zone!"<<endl;
+    res=RCode::NoError;
+    return true;
+  }
 
   string nsdomain(qname);
 
@@ -165,7 +184,7 @@ bool SyncRes::doOOBResolve(const string &qname, const QType &qtype, vector<DNSRe
   else 
     res=0;
 
-  return true;;
+  return true;
 }
 
 int SyncRes::doResolve(const string &qname, const QType &qtype, vector<DNSResourceRecord>&ret, int depth, set<GetBestNSAnswer>& beenthere)
@@ -178,6 +197,34 @@ int SyncRes::doResolve(const string &qname, const QType &qtype, vector<DNSResour
   
   int res=0;
   if(!(d_nocache && qtype.getCode()==QType::NS && qname==".")) {
+    if(d_cacheonly) { // very limited OOB support
+      LWResult lwr;
+      LOG<<prefix<<qname<<": Recursion not requested for '"<<qname<<"|"<<qtype.getName()<<"', peeking at auth/forward zones"<<endl;
+      string authname(qname);
+      domainmap_t::const_iterator iter=getBestAuthZone(&authname);
+      if(iter != s_domainmap.end()) {
+	const vector<ComboAddress>& servers = iter->second.d_servers;
+	if(servers.empty()) {
+	  ret.clear();
+	  doOOBResolve(qname, qtype, ret, depth, res);
+	  return res;
+	}
+	else {
+	  const ComboAddress remoteIP = servers.front();
+	  LOG<<prefix<<qname<<": forwarding query to hardcoded nameserver '"<< remoteIP.toStringWithPort()<<"' for zone '"<<authname<<"'"<<endl;
+
+	  res=asyncresolve(remoteIP, qname, qtype.getCode(), false, false, &d_now, &lwr);    
+	  // filter out the good stuff from lwr.result()
+
+	  for(LWResult::res_t::const_iterator i=lwr.d_result.begin();i!=lwr.d_result.end();++i) {
+	    if(i->d_place == DNSResourceRecord::ANSWER)
+	      ret.push_back(*i);
+	  }
+	  return res;
+	}
+      }
+    }
+
     if(doCNAMECacheCheck(qname,qtype,ret,depth,res)) // will reroute us if needed
       return res;
     
@@ -187,14 +234,15 @@ int SyncRes::doResolve(const string &qname, const QType &qtype, vector<DNSResour
 
   if(d_cacheonly)
     return 0;
-
+    
   LOG<<prefix<<qname<<": No cache hit for '"<<qname<<"|"<<qtype.getName()<<"', trying to find an appropriate NS record"<<endl;
 
   string subdomain(qname);
 
   set<string, CIStringCompare> nsset;
+  bool flawedNSSet=false;
   for(int tries=0;tries<2 && nsset.empty();++tries) {
-    subdomain=getBestNSNamesFromCache(subdomain,nsset,depth, beenthere); //  pass beenthere to both occasions
+    subdomain=getBestNSNamesFromCache(subdomain, nsset, &flawedNSSet, depth, beenthere); //  pass beenthere to both occasions
 
     if(nsset.empty()) { // must've lost root records
       LOG<<prefix<<qname<<": our root expired, repriming from hints and retrying"<<endl;
@@ -202,7 +250,7 @@ int SyncRes::doResolve(const string &qname, const QType &qtype, vector<DNSResour
     }
   }
 
-  if(!(res=doResolveAt(nsset, subdomain, qname, qtype, ret, depth, beenthere)))
+  if(!(res=doResolveAt(nsset, subdomain, flawedNSSet, qname, qtype, ret, depth, beenthere)))
     return 0;
   
   LOG<<prefix<<qname<<": failed (res="<<res<<")"<<endl;
@@ -220,35 +268,24 @@ vector<ComboAddress> SyncRes::getAs(const string &qname, int depth, set<GetBestN
   typedef vector<ComboAddress> ret_t;
   ret_t ret;
 
-  if(!doResolve(qname,QType(QType::A), res,depth+1,beenthere) && !res.empty()) {
+  if(!doResolve(qname, s_doIPv6 ? QType(QType::ADDR) : QType(QType::A), res,depth+1,beenthere) && !res.empty()) {  // this consults cache, OR goes out
     for(res_t::const_iterator i=res.begin(); i!= res.end(); ++i) {
-      if(i->qtype.getCode()==QType::A) {
+      if(i->qtype.getCode()==QType::A || i->qtype.getCode()==QType::AAAA) {
 	ret.push_back(ComboAddress(i->content, 53));
       }
     }
   }
-
-  if(s_doIPv6) {
-    typedef set<DNSResourceRecord> ipv6_t;
-    ipv6_t ipv6;
-    if(RC.get(d_now.tv_sec, qname, QType(QType::AAAA), &ipv6) > 0) {
-      for(ipv6_t::const_iterator i=ipv6.begin(); i != ipv6.end(); ++i) 
-	ret.push_back(ComboAddress(i->content, 53));
-    }
-  }
   
   if(ret.size() > 1) {
-    random_shuffle(ret.begin(), ret.end());
+    random_shuffle(ret.begin(), ret.end(), dns_random);
 
-    // move 'best' address up front
-    nsspeeds_t::iterator best=s_nsSpeeds.find(qname);
+    // move 'best' address for this nameserver name up front
+    nsspeeds_t::iterator best = s_nsSpeeds.find(qname);  
 
     if(best != s_nsSpeeds.end())
-      for(ret_t::iterator i=ret.begin(); i != ret.end(); ++i) {
-	//	cerr<<"Is "<<i->toString()<<" equal to "<<best->second.d_best.toString()<<"?\n";
-	if(*i==best->second.d_best) {
+      for(ret_t::iterator i=ret.begin(); i != ret.end(); ++i) {  
+	if(*i==best->second.d_best) {  // got the fastest one
 	  if(i!=ret.begin()) {
-	    //	    cerr<<"Moving "<<best->second.d_best.toString()<<" up front!\n";
 	    *i=*ret.begin();
 	    *ret.begin()=best->second.d_best;
 	  }
@@ -260,7 +297,7 @@ vector<ComboAddress> SyncRes::getAs(const string &qname, int depth, set<GetBestN
   return ret;
 }
 
-void SyncRes::getBestNSFromCache(const string &qname, set<DNSResourceRecord>&bestns, int depth, set<GetBestNSAnswer>& beenthere)
+void SyncRes::getBestNSFromCache(const string &qname, set<DNSResourceRecord>&bestns, bool* flawedNSSet, int depth, set<GetBestNSAnswer>& beenthere)
 {
   string prefix, subdomain(qname);
   if(s_log) {
@@ -272,27 +309,31 @@ void SyncRes::getBestNSFromCache(const string &qname, set<DNSResourceRecord>&bes
   do {
     LOG<<prefix<<qname<<": Checking if we have NS in cache for '"<<subdomain<<"'"<<endl;
     set<DNSResourceRecord> ns;
+    *flawedNSSet = false;
     if(RC.get(d_now.tv_sec, subdomain, QType(QType::NS), &ns) > 0) {
       for(set<DNSResourceRecord>::const_iterator k=ns.begin();k!=ns.end();++k) {
 	if(k->ttl > (unsigned int)d_now.tv_sec ) { 
-	  set<DNSResourceRecord>aset;
+	  set<DNSResourceRecord> aset;
 
 	  DNSResourceRecord rr=*k;
 	  rr.content=k->content;
-	  if(!dottedEndsOn(rr.content, subdomain) || RC.get(d_now.tv_sec, rr.content, QType(QType::A), s_log ? &aset : 0) > 5) {
+	  if(!dottedEndsOn(rr.content, subdomain) || RC.get(d_now.tv_sec, rr.content, s_doIPv6 ? QType(QType::ADDR) : QType(QType::A),
+							    s_log ? &aset : 0) > 5) {
 	    bestns.insert(rr);
 	    
 	    LOG<<prefix<<qname<<": NS (with ip, or non-glue) in cache for '"<<subdomain<<"' -> '"<<rr.content<<"'"<<endl;
 	    LOG<<prefix<<qname<<": within bailiwick: "<<dottedEndsOn(rr.content, subdomain);
 	    if(!aset.empty()) {
-	      LOG<<", in cache, ttl="<<(unsigned int)(((time_t)aset.begin()->ttl- d_now.tv_sec ))<<endl;
+	      LOG<<",  in cache, ttl="<<(unsigned int)(((time_t)aset.begin()->ttl- d_now.tv_sec ))<<endl;
 	    }
 	    else {
 	      LOG<<", not in cache / did not look at cache"<<endl;
 	    }
 	  }
-	  else
+	  else {
+	    *flawedNSSet=true;
 	    LOG<<prefix<<qname<<": NS in cache for '"<<subdomain<<"', but needs glue ("<<k->content<<") which we miss or is expired"<<endl;
+	  }
 	}
       }
       if(!bestns.empty()) {
@@ -300,13 +341,14 @@ void SyncRes::getBestNSFromCache(const string &qname, set<DNSResourceRecord>&bes
 	answer.qname=qname; answer.bestns=bestns;
 	if(beenthere.count(answer)) {
 	  LOG<<prefix<<qname<<": We have NS in cache for '"<<subdomain<<"' but part of LOOP! Trying less specific NS"<<endl;
-	  for( set<GetBestNSAnswer>::const_iterator j=beenthere.begin();j!=beenthere.end();++j)
-	    LOG<<prefix<<qname<<": beenthere: "<<j->qname<<" ("<<(unsigned int)j->bestns.size()<<")"<<endl;
+	  if(s_log)
+	    for( set<GetBestNSAnswer>::const_iterator j=beenthere.begin();j!=beenthere.end();++j)
+	      LOG<<prefix<<qname<<": beenthere: "<<j->qname<<" ("<<(unsigned int)j->bestns.size()<<")"<<endl;
 	  bestns.clear();
 	}
 	else {
 	  beenthere.insert(answer);
-	  LOG<<prefix<<qname<<": We have NS in cache for '"<<subdomain<<"'"<<endl;
+	  LOG<<prefix<<qname<<": We have NS in cache for '"<<subdomain<<"' (flawedNSSet="<<*flawedNSSet<<")"<<endl;
 	  return;
 	}
       }
@@ -327,20 +369,25 @@ SyncRes::domainmap_t::const_iterator SyncRes::getBestAuthZone(string* qname)
 }
 
 /** doesn't actually do the work, leaves that to getBestNSFromCache */
-string SyncRes::getBestNSNamesFromCache(const string &qname,set<string, CIStringCompare>& nsset, int depth, set<GetBestNSAnswer>&beenthere)
+string SyncRes::getBestNSNamesFromCache(const string &qname, set<string, CIStringCompare>& nsset, bool* flawedNSSet, int depth, set<GetBestNSAnswer>&beenthere)
 {
   string subdomain(qname);
-
   string authdomain(qname);
   
   domainmap_t::const_iterator iter=getBestAuthZone(&authdomain);
   if(iter!=s_domainmap.end()) {
-    nsset.insert(iter->second.d_server);
+    if( iter->second.d_servers.empty() )
+      nsset.insert(string()); // this gets picked up in doResolveAt, if empty it means "we are auth", otherwise it denotes a forward
+    else {
+      for(vector<ComboAddress>::const_iterator server=iter->second.d_servers.begin(); server != iter->second.d_servers.end(); ++server)
+	nsset.insert(server->toStringWithPort());
+    }
+
     return authdomain;
   }
 
   set<DNSResourceRecord> bestns;
-  getBestNSFromCache(subdomain, bestns, depth, beenthere);
+  getBestNSFromCache(subdomain, bestns, flawedNSSet, depth, beenthere);
 
   for(set<DNSResourceRecord>::const_iterator k=bestns.begin();k!=bestns.end();++k) {
     nsset.insert(k->content);
@@ -501,7 +548,7 @@ inline vector<string> SyncRes::shuffleInSpeedOrder(set<string, CIStringCompare>
     rnameservers.push_back(*i);
     speeds[*i]=s_nsSpeeds[*i].get(&d_now);
   }
-  random_shuffle(rnameservers.begin(),rnameservers.end());
+  random_shuffle(rnameservers.begin(),rnameservers.end(), dns_random);
   speedOrder so(speeds);
   stable_sort(rnameservers.begin(),rnameservers.end(), so);
   
@@ -534,10 +581,15 @@ struct TCacheComp
   }
 };
 
-
+static bool magicAddrMatch(const QType& query, const QType& answer)
+{
+  if(query.getCode() != QType::ADDR)
+    return false;
+  return answer.getCode() == QType::A || answer.getCode() == QType::AAAA;
+}
 
 /** returns -1 in case of no results, rcode otherwise */
-int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth, const string &qname, const QType &qtype, 
+int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth, bool flawedNSSet, const string &qname, const QType &qtype, 
 			 vector<DNSResourceRecord>&ret, 
 			 int depth, set<GetBestNSAnswer>&beenthere)
 {
@@ -547,18 +599,19 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
     prefix.append(depth, ' ');
   }
   
-  LWRes::res_t result;
-
   LOG<<prefix<<qname<<": Cache consultations done, have "<<(unsigned int)nameservers.size()<<" NS to contact"<<endl;
 
   for(;;) { // we may get more specific nameservers
-    result.clear();
-
     vector<string> rnameservers=shuffleInSpeedOrder(nameservers, s_log ? (prefix+qname+": ") : string() );
 
     for(vector<string>::const_iterator tns=rnameservers.begin();;++tns) { 
       if(tns==rnameservers.end()) {
-	LOG<<prefix<<qname<<": Failed to resolve via any of the "<<(unsigned int)rnameservers.size()<<" offered NS"<<endl;
+	LOG<<prefix<<qname<<": Failed to resolve via any of the "<<(unsigned int)rnameservers.size()<<" offered NS at level '"<<auth<<"'"<<endl;
+	if(auth!="." && flawedNSSet) {
+	  g_stats.nsSetInvalidations++;
+	  LOG<<prefix<<qname<<": Invalidating nameservers for level '"<<auth<<"', next query might succeed"<<endl;
+	  RC.doWipeCache(auth, QType::NS);
+	}
 	return -1;
       }
       if(qname==*tns && qtype.getCode()==QType::A) {
@@ -571,24 +624,35 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
       remoteIPs_t::const_iterator remoteIP;
       bool doTCP=false;
       int resolveret;
+      bool pierceDontQuery=false;
 
+      LWResult lwr;
       if(tns->empty()) {
 	LOG<<prefix<<qname<<": Domain is out-of-band"<<endl;
-	doOOBResolve(qname, qtype, result, depth, d_lwr.d_rcode);
-	d_lwr.d_tcbit=false;
-	d_lwr.d_aabit=true;
+	doOOBResolve(qname, qtype, lwr.d_result, depth, lwr.d_rcode);
+	lwr.d_tcbit=false;
+	lwr.d_aabit=true;
       }
       else {
 	LOG<<prefix<<qname<<": Trying to resolve NS '"<<*tns<<"' ("<<1+tns-rnameservers.begin()<<"/"<<(unsigned int)rnameservers.size()<<")"<<endl;
+
 	if(!isCanonical(*tns)) {
-	  LOG<<prefix<<qname<<": Domain has hardcoded nameserver"<<endl;
-	  remoteIPs.push_back(ComboAddress(*tns, 53));
+	  LOG<<prefix<<qname<<": Domain has hardcoded nameserver(s)"<<endl;
+
+	  pair<string,string> ipport=splitField(*tns, ':');
+	  ComboAddress addr(ipport.first, ipport.second.empty() ? 53 : lexical_cast<uint16_t>(ipport.second));
+	  
+	  remoteIPs.push_back(addr);
+	  pierceDontQuery=true;
 	}
-	else
+	else {
 	  remoteIPs=getAs(*tns, depth+1, beenthere);
+	  pierceDontQuery=false;
+	}
 
 	if(remoteIPs.empty()) {
 	  LOG<<prefix<<qname<<": Failed to get IP for NS "<<*tns<<", trying next if available"<<endl;
+	  flawedNSSet=true;
 	  continue;
 	}
 	else {
@@ -601,22 +665,31 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
 	  LOG<<endl;
 
 	}
+
 	for(remoteIP = remoteIPs.begin(); remoteIP != remoteIPs.end(); ++remoteIP) {
-	  LOG<<prefix<<qname<<": Trying IP "<< remoteIP->toString() <<", asking '"<<qname<<"|"<<qtype.getName()<<"'"<<endl;
+	  LOG<<prefix<<qname<<": Trying IP "<< remoteIP->toStringWithPort() <<", asking '"<<qname<<"|"<<qtype.getName()<<"'"<<endl;
+	  extern NetmaskGroup* g_dontQuery;
 	  
 	  if(s_throttle.shouldThrottle(d_now.tv_sec, make_tuple(*remoteIP, qname, qtype.getCode()))) {
 	    LOG<<prefix<<qname<<": query throttled "<<endl;
 	    s_throttledqueries++; d_throttledqueries++;
 	    continue;
+	  } 
+	  else if(!pierceDontQuery && g_dontQuery && g_dontQuery->match(&*remoteIP)) {
+	    LOG<<prefix<<qname<<": not sending query to " << remoteIP->toString() << ", blocked by 'dont-query' setting" << endl;
+	    continue;
 	  }
 	  else {
 	    s_outqueries++; d_outqueries++;
 	  TryTCP:
 	    if(doTCP) {
+	      LOG<<prefix<<qname<<": using TCP with "<< remoteIP->toStringWithPort() <<endl;
 	      s_tcpoutqueries++; d_tcpoutqueries++;
 	    }
 	    
-	    resolveret=d_lwr.asyncresolve(*remoteIP, qname, qtype.getCode(), doTCP, &d_now);    // <- we go out on the wire!
+	    resolveret=asyncresolve(*remoteIP, qname, 
+				    (qtype.getCode() == QType::ADDR ? QType::ANY : qtype.getCode()), 
+				    doTCP, d_doEDNS0, &d_now, &lwr);    // <- we go out on the wire!
 	    if(resolveret != 1) {
 	      if(resolveret==0) {
 		LOG<<prefix<<qname<<": timeout resolving "<< (doTCP ? "over TCP" : "")<<endl;
@@ -624,12 +697,12 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
 		s_outgoingtimeouts++;
 	      }
 	      else if(resolveret==-2) {
-		LOG<<prefix<<qname<<": hit a local resource limit resolving "<< (doTCP ? "over TCP" : "")<<endl;
+		LOG<<prefix<<qname<<": hit a local resource limit resolving"<< (doTCP ? " over TCP" : "")<<", probable error: "<<stringerror()<<endl;
 		g_stats.resourceLimits++;
 	      }
 	      else {
 		s_unreachables++; d_unreachables++;
-		LOG<<prefix<<qname<<": error resolving "<< (doTCP ? "over TCP" : "") << endl;
+		LOG<<prefix<<qname<<": error resolving"<< (doTCP ? " over TCP" : "") <<", possible error: "<<strerror(errno)<< endl;
 	      }
 	      
 	      if(resolveret!=-2) { // don't account for resource limits, they are our own fault
@@ -652,9 +725,7 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
 	if(remoteIP == remoteIPs.end())  // we tried all IP addresses, none worked
 	  continue; 
 	
-	result=d_lwr.result();
-      
-	if(d_lwr.d_tcbit) {
+	if(lwr.d_tcbit) {
 	  if(!doTCP) {
 	    doTCP=true;
 	    LOG<<prefix<<qname<<": truncated bit set, retrying via TCP"<<endl;
@@ -664,26 +735,30 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
 	  return RCode::ServFail;
 	}
 	
-	if(d_lwr.d_rcode==RCode::ServFail) {
+	if(lwr.d_rcode==RCode::ServFail) {
 	  LOG<<prefix<<qname<<": "<<*tns<<" returned a ServFail, trying sibling IP or NS"<<endl;
 	  s_throttle.throttle(d_now.tv_sec,make_tuple(*remoteIP, qname, qtype.getCode()),60,3);
 	  continue;
 	}
-	LOG<<prefix<<qname<<": Got "<<(unsigned int)result.size()<<" answers from "<<*tns<<" ("<< remoteIP->toString() <<"), rcode="<<d_lwr.d_rcode<<", in "<<d_lwr.d_usec/1000<<"ms"<<endl;
+	LOG<<prefix<<qname<<": Got "<<(unsigned int)lwr.d_result.size()<<" answers from "<<*tns<<" ("<< remoteIP->toString() <<"), rcode="<<lwr.d_rcode<<", in "<<lwr.d_usec/1000<<"ms"<<endl;
 
 	/*  // for you IPv6 fanatics :-)
 	if(remoteIP->sin4.sin_family==AF_INET6)
-	  d_lwr.d_usec/=3;
+	  lwr.d_usec/=3;
 	*/
 
-	s_nsSpeeds[*tns].submit(*remoteIP, d_lwr.d_usec, &d_now);
+	s_nsSpeeds[*tns].submit(*remoteIP, lwr.d_usec, &d_now);
       }
 
       typedef map<pair<string, QType>, set<DNSResourceRecord>, TCacheComp > tcache_t;
       tcache_t tcache;
 
       // reap all answers from this packet that are acceptable
-      for(LWRes::res_t::const_iterator i=result.begin();i!=result.end();++i) {
+      for(LWResult::res_t::const_iterator i=lwr.d_result.begin();i != lwr.d_result.end();++i) {
+	if(i->qtype.getCode() == QType::OPT) {
+	  LOG<<prefix<<qname<<": skipping OPT answer '"<<i->qname<<"' from '"<<auth<<"' nameservers" <<endl;
+	  continue;
+	}
 	LOG<<prefix<<qname<<": accept answer '"<<i->qname<<"|"<<i->qtype.getName()<<"|"<<i->content<<"' from '"<<auth<<"' nameservers? ";
 	if(i->qtype.getCode()==QType::ANY) {
 	  LOG<<"NO! - we don't accept 'ANY' data"<<endl;
@@ -691,7 +766,7 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
 	}
 	  
 	if(dottedEndsOn(i->qname, auth)) {
-	  if(d_lwr.d_aabit && d_lwr.d_rcode==RCode::NoError && i->d_place==DNSResourceRecord::ANSWER && ::arg().contains("delegation-only",auth)) {
+	  if(lwr.d_aabit && lwr.d_rcode==RCode::NoError && i->d_place==DNSResourceRecord::ANSWER && ::arg().contains("delegation-only",auth)) {
 	    LOG<<"NO! Is from delegation-only zone"<<endl;
 	    s_nodelegated++;
 	    return RCode::NXDomain;
@@ -725,7 +800,7 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
 	    ((tcache_t::value_type::second_type::value_type*)&(*j))->ttl=lowestTTL;
 	}
 
-	RC.replace(d_now.tv_sec, i->first.first, i->first.second, i->second, d_lwr.d_aabit);
+	RC.replace(d_now.tv_sec, i->first.first, i->first.second, i->second, lwr.d_aabit);
       }
       set<string, CIStringCompare> nsset;  
       LOG<<prefix<<qname<<": determining status after receiving this packet"<<endl;
@@ -733,9 +808,9 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
       bool done=false, realreferral=false, negindic=false;
       string newauth, soaname, newtarget;
 
-      for(LWRes::res_t::const_iterator i=result.begin();i!=result.end();++i) {
+      for(LWResult::res_t::const_iterator i=lwr.d_result.begin();i!=lwr.d_result.end();++i) {
 	if(i->d_place==DNSResourceRecord::AUTHORITY && dottedEndsOn(qname,i->qname) && i->qtype.getCode()==QType::SOA && 
-	   d_lwr.d_rcode==RCode::NXDomain) {
+	   lwr.d_rcode==RCode::NXDomain) {
 	  LOG<<prefix<<qname<<": got negative caching indication for RECORD '"<<qname+"'"<<endl;
 	  ret.push_back(*i);
 
@@ -749,18 +824,20 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
 	  replacing_insert(s_negcache, ne);
 	  negindic=true;
 	}
-	else if(i->d_place==DNSResourceRecord::ANSWER && i->qname==qname && i->qtype.getCode()==QType::CNAME && (!(qtype==QType(QType::CNAME)))) {
+	else if(i->d_place==DNSResourceRecord::ANSWER && iequals(i->qname, qname) && i->qtype.getCode()==QType::CNAME && (!(qtype==QType(QType::CNAME)))) {
 	  ret.push_back(*i);
 	  newtarget=i->content;
 	}
 	// for ANY answers we *must* have an authoritive answer
 	else if(i->d_place==DNSResourceRecord::ANSWER && !Utility::strcasecmp(i->qname.c_str(),qname.c_str()) && 
-		( (i->qtype==qtype) ||
-		                      ( qtype==QType(QType::ANY) && d_lwr.d_aabit)))  {
+		(
+		 i->qtype==qtype || (lwr.d_aabit && (qtype==QType(QType::ANY) || magicAddrMatch(qtype, i->qtype) ) )
+		) 
+	       )   
+	  {
 	  
 	  LOG<<prefix<<qname<<": answer is in: resolved to '"<< i->content<<"|"<<i->qtype.getName()<<"'"<<endl;
 
-
 	  done=true;
 	  ret.push_back(*i);
 	}
@@ -774,8 +851,8 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
 	    LOG<<prefix<<qname<<": got upwards/level NS record '"<<i->qname<<"' -> '"<<i->content<<"', had '"<<auth<<"'"<<endl;
 	  nsset.insert(i->content);
 	}
-	else if(i->d_place==DNSResourceRecord::AUTHORITY && dottedEndsOn(qname,i->qname) && i->qtype.getCode()==QType::SOA && 
-	   d_lwr.d_rcode==RCode::NoError) {
+	else if(!done && i->d_place==DNSResourceRecord::AUTHORITY && dottedEndsOn(qname,i->qname) && i->qtype.getCode()==QType::SOA && 
+	   lwr.d_rcode==RCode::NoError) {
 	  LOG<<prefix<<qname<<": got negative caching indication for '"<< (qname+"|"+i->qtype.getName()+"'") <<endl;
 	  ret.push_back(*i);
 	  
@@ -795,16 +872,25 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
 	LOG<<prefix<<qname<<": status=got results, this level of recursion done"<<endl;
 	return 0;
       }
-      if(d_lwr.d_rcode==RCode::NXDomain) {
+      if(lwr.d_rcode==RCode::NXDomain) {
 	LOG<<prefix<<qname<<": status=NXDOMAIN, we are done "<<(negindic ? "(have negative SOA)" : "")<<endl;
 	return RCode::NXDomain;
       }
       if(!newtarget.empty()) {
+	if(iequals(newtarget,qname)) {
+	  LOG<<prefix<<qname<<": status=got a CNAME referral to self, returning SERVFAIL"<<endl;
+	  return RCode::ServFail;
+	}
+	if(depth > 10) {
+	  LOG<<prefix<<qname<<": status=got a CNAME referral, but recursing too deep, returning SERVFAIL"<<endl;
+	  return RCode::ServFail;
+	}
 	LOG<<prefix<<qname<<": status=got a CNAME referral, starting over with "<<newtarget<<endl;
+
 	set<GetBestNSAnswer>beenthere2;
-	return doResolve(newtarget, qtype, ret,0,beenthere2);
+	return doResolve(newtarget, qtype, ret, depth + 1, beenthere2);
       }
-      if(nsset.empty() && !d_lwr.d_rcode) {
+      if(nsset.empty() && !lwr.d_rcode) {
 	LOG<<prefix<<qname<<": status=noerror, other types may exist, but we are done "<<(negindic ? "(have negative SOA)" : "")<<endl;
 	return 0;
       }
@@ -815,7 +901,7 @@ int SyncRes::doResolveAt(set<string, CIStringCompare> nameservers, string auth,
 	break; 
       }
       else if(isCanonical(*tns)) {
-	goto wasLame;;
+	goto wasLame;
       }
     }
   }
@@ -859,9 +945,7 @@ void SyncRes::addCruft(const string &qname, vector<DNSResourceRecord>& ret)
 	host=string(k->content.c_str() + fields[3].first, fields[3].second - fields[3].first);
       else 
 	continue;
-      doResolve(host, QType(QType::A), addit, 1, beenthere);
-      if(*l_doIPv6AP)
-	doResolve(host, QType(QType::AAAA), addit, 1, beenthere);
+      doResolve(host, *l_doIPv6AP ? QType(QType::ADDR) : QType(QType::A), addit, 1, beenthere);
     }
   
   sort(addit.begin(), addit.end());
@@ -878,8 +962,9 @@ void SyncRes::addCruft(const string &qname, vector<DNSResourceRecord>& ret)
 void SyncRes::addAuthorityRecords(const string& qname, vector<DNSResourceRecord>& ret, int depth)
 {
   set<DNSResourceRecord> bestns;
-  set<GetBestNSAnswer>beenthere;
-  getBestNSFromCache(qname, bestns, depth,beenthere);
+  set<GetBestNSAnswer> beenthere;
+  bool dontcare;
+  getBestNSFromCache(qname, bestns, &dontcare, depth, beenthere);
 
   for(set<DNSResourceRecord>::const_iterator k=bestns.begin();k!=bestns.end();++k) {
     DNSResourceRecord ns=*k;
